#ifndef _FLITVIEW_H_
#define _FLITVIEW_H_

#include "list.h"

#define FLITVIEW_VER "flitview v0.5 2009-02-02"

#define USAGE \
 "usage: flitview [logfile]\n"\
 "  without logfile, flitview get log data from standard input.\n"\
 "  visualize network state using ncurses library.\n"\
 " options\n"\
 "    --help: show this message\n"\
 "    --version: show version number\n"

typedef unsigned int uint;

typedef struct {
    uint clk, phase;
    uint y, x;
    uint direction;
    uint site;
    uint data;
} log_t;

typedef struct {
    log_t log;
    struct list_head list;
} loglist_t;

static char *fname = NULL;

#define NUM_DIRECTION 5


static const uint coordinates[NUM_DIRECTION][3][2] = {
    {{1, 5}, {2, 5}, {1, 8}},   // N
    {{3, 11}, {3, 8}, {4, 11}}, // E
    {{6, 8}, {5, 8}, {6, 5}},   // S
    {{4, 1}, {4, 4}, {3, 1}},   // W
    {{1, 1}, {1, 1}, {1, 1}},   // C
};

static const char *phasetab[] = {
    "update   ",
    "switching",
    "comm     ",
};

class flitview {
    log_t *log;
    uint logsize;
    char simver[256];
    FILE *fp;
    int term_lines, term_columns;
    WINDOW ***win;
    uint clk, phase;
    uint current;
    uint maxclk;
    char str[256];
    uint noc_col, noc_row;
    
    void search_next();
    void search_prev();
    void phase_jump(int num);
    void goto_cycle(uint clk);
    void clear();
    void draw();
    void update();
    void cursesgets();
    void log_search(uint *n, uint clk, uint phase);
    int compare(log_t *current, uint clk, uint phase);
    void initscreen();
public:
    char *fname;

    void read_log();
    void show_cui();
    flitview();
    ~flitview();
};

#endif
