#include "network.h"

/**********************************************************************/
/**
 * Bitmap constructor
 * /param words the bitmap size (words * 32 bit)
 */
Bitmap::Bitmap(unsigned int words) {
    this->words = words;
    this->nbits = words*sizeof(unsigned int);

    bitmap = new unsigned int [words];
    memset(bitmap, 0, words*sizeof(unsigned int));
}

/**********************************************************************/
/**
 * Bitmap destructor
 */
Bitmap::~Bitmap() {
//    delete bitmap;
}

/**********************************************************************/
/**
 * check whether all bits are zero
 * \return 0: 1:
 */
int Bitmap::isclear() {
    for(int i = 0; i < words; i++)
        if(bitmap[i]) 
            return 0;
    return 1;
}

/**********************************************************************/
/**
 * set the n th bit
 */
void Bitmap::set(unsigned int n) {
    bitmap[n >> 5] |= (1 << (n & 0x1f));
}

/**********************************************************************/
/**
 * reset the n th bit
 */
void Bitmap::reset(unsigned int n) {
    bitmap[n >> 5] &= ~(1 << (n & 0x1f));
}

/**********************************************************************/
/**
 * find LSB from bit map (unsigned int [])
 * \return LSB
 */
int Bitmap::lsb() {
    return lsb(bitmap, words);
}

/**********************************************************************/
/**
 * find MSB from bit map (unsigned int [])
 * \return MSB
 */
int Bitmap::msb() {
    return msb(bitmap, words);
}

/**********************************************************************/
/**
 * find LSB from bit map
 * \param bitmap the target bit map
 * \param words the size (word) of the bitmap
 * \return LSB
 */
int Bitmap::lsb(unsigned int *bitmap, unsigned int words) {
    for(unsigned int i = 0; i < words; i++)
        if(bitmap[i])
            return lsb(bitmap[i]) | (i << 5);
    return NOT_FOUND;
}

/**********************************************************************/
/**
 * find MSB from bit map
 * \param bitmap the target bit map
 * \param words the size (word) of the bitmap
 * \return MSB
 */
int Bitmap::msb(unsigned int *bitmap, unsigned int words) {
    for(unsigned int i = words - 1; i >= 0; i--)
        if(bitmap[i])
            return msb(bitmap[i]) | (i << 5);
    return NOT_FOUND;
}

/**********************************************************************/
/**
 * find LSB from a word
 * \param x the target word
 * \return LSB
 */
int Bitmap::lsb(unsigned int x) {
    if(x == 0) return NOT_FOUND;
    unsigned int n = 0;
    if((x & 0x0000ffff) == 0) n |= 16, x >>= 16;
    if((x & 0x000000ff) == 0) n |= 8, x >>= 8;
    if((x & 0x0000000f) == 0) n |= 4, x >>= 4;
    if((x & 0x00000003) == 0) n |= 2, x >>= 2;
    if((x & 0x00000001) == 0) n |= 1;
    return n;
}

/**********************************************************************/
/**
 * find MSB from a word
 * \param x the target word
 * \return MSB
 */
int Bitmap::msb(unsigned int x) {
    if(x == 0) return NOT_FOUND;
    unsigned int n = 0;
    if((x & 0xffff0000) == 0) n |= 16, x <<= 16;
    if((x & 0xff000000) == 0) n |= 8, x <<= 8;
    if((x & 0xf0000000) == 0) n |= 4, x <<= 4;
    if((x & 0xc0000000) == 0) n |= 2, x <<= 2;
    if((x & 0x80000000) == 0) n |= 1;
    return n;
}

/**********************************************************************/
/**
 * print bitmap
 */
void Bitmap::print() {
    for(int i = 0; i < words; i++)
        printf("%08x", bitmap[i]);
    putchar('\n');
}
