/**********************************************************************
 * SimMips: Simple Computer Simulator of MIPS    Arch Lab. TOKYO TECH *
 **********************************************************************/
/**********************************************************************/
/* WARNING!! This file contains the code of Simplescalar/PISA.        */
/*           So, please handle it with care.                          */
/**********************************************************************/
/* WARNING!! This may contain some bugs. Please fix me.               */
/**********************************************************************/
#include "define.h"
#ifndef CYGWIN
#ifndef FreeBSD
#include <sgtty.h>
#endif
#include <sys/syscall.h>
#include <sys/dir.h>
#endif
#include <errno.h>
#include <time.h>
#include <setjmp.h>
#include <limits.h>
#include <utime.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/utsname.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/ioctl.h>
#include <sys/select.h>

/**********************************************************************/
#define SYSCALL_BUFSIZE  256
#define NUM_IOCTL_BYTES  128
#define ROUND_UP(a) ((a + (PAGE_SIZE - 1)) & ~(PAGE_SIZE - 1))

typedef int (*Printf)(const char *format, ...);

int silentprintf (const char *format, ...) {
    return 0;
}

void fetal(const char *msg) {
    fprintf(stderr, "%s", msg);
    exit(-1);
}

void memstrcpy(MainMemory *mem, uint032_t addr, uint008_t *str) {
    int i;
    for(i = 0; i < SYSCALL_BUFSIZE; i++) {
        mem->read1b(addr + i, str + i);
        if(str[i] == '\0') break;
    }
    if(i == SYSCALL_BUFSIZE) 
        fetal("## SYSCALL filename too long.\n");
}

/* System Call Definition *********************************************/
/**********************************************************************/
enum {
    //// asm/unistd.h
    SYS_EXIT         = 4001,
    SYS_READ         = 4003,
    SYS_WRITE        = 4004,
    SYS_OPEN         = 4005,
    SYS_CLOSE        = 4006,    
    SYS_TIME         = 4013,
    SYS_LSEEK        = 4019,
    SYS_GETPID       = 4020,
    SYS_KILL         = 4037,
    SYS_RMDIR        = 4040,
    SYS_TIMES        = 4043,
    SYS_UTIME        = 4030,
    SYS_BRK          = 4045,
    SYS_IOCTL        = 4054,
    SYS_SETRLIMIT    = 4075,
    SYS_GETRLIMIT    = 4076,
    SYS_GETRUSAGE    = 4077,
    SYS_MMAP         = 4090,
    SYS_MUNMAP       = 4091,
    SYS_FTRUNCATE    = 4093,
    SYS_STAT         = 4106,
    SYS_FSTAT        = 4108,
    SYS_UNAME        = 4122,
    SYS_WRITEV       = 4146,
    SYS_MREMAP       = 4167,
    SYS_RT_SIGACTION = 4194,
    SYS_STAT64       = 4213,
    SYS_FUTEX        = 4238,
    SYS_FORK         = 4002,
    SYS_CREATE       = 4008,
    SYS_UNLINK       = 4010,
    SYS_CHDIR        = 4012,
    SYS_CHMOD        = 4015,
    SYS_CHOWN        = 4202,
    SYS_GETUID       = 4024,
    SYS_GETGID       = 4047,
    SYS_ACCESS       = 4033,
    SYS_LSTAT        = 4107,
    SYS_DUP          = 4041,
    SYS_DUP2         = 4063,
    SYS_PIPE         = 4042,
    SYS_FCNTL        = 4055,
    SYS_GETEUID      = 4049,
    SYS_GETEGID      = 4050,
    SYS_GETTIMEOFDAY = 4078,
    SYS_GETDENTS     = 4141,
    SYS_RT_SIGPROCMASK = 4195,
    SYS_GETCWD       = 4203,

    //// bits/fcntl.h
    MIPS_O_RDONLY    = 0x0000,
    MIPS_O_WRONLY    = 0x0001,
    MIPS_O_RDWR      = 0x0002,
    MIPS_O_ACCMODE   = 0x0003,
    MIPS_O_APPEND    = 0x0008,
    MIPS_O_ASYNC     = 0x1000,
    MIPS_O_CREAT     = 0x0100,
    MIPS_O_EXCL      = 0x0400,
    MIPS_O_NOCTTY    = 0x0800,
    MIPS_O_NONBLOCK  = 0x0080,
    MIPS_O_SYNC      = 0x0010,
    MIPS_O_TRUNC     = 0x0200,
    MIPS_O_DIRECTORY = 0x10000,

    //// bits/mman.h
    MIPS_MAP_FIXED      = 0x0010,
    MIPS_MAP_ANONYMOUS  = 0x0800,
    MIPS_MAP_GROWSDOWN  = 0x1000,
    MIPS_MREMAP_MAYMOVE = 1,
    MIPS_MREMAP_FIXED   = 2,
    MIPS_MAP_FAILED     = -1,

    SS_IOCTL_TIOCGETP	= 0x7408,
    SS_IOCTL_TIOCSETP	= 0x7409,
    SS_IOCTL_TCGETP		= 1076130901,
    SS_IOCTL_TCGETA		= 1075082331,
    SS_IOCTL_TIOCGLTC	= 0x7474,
    SS_IOCTL_TIOCSLTC	= 0x7475,
    SS_IOCTL_TIOCGWINSZ	= 1074295912,
    SS_IOCTL_TCSETAW	   = 0x5403,
    SS_IOCTL_TIOCGETC	= 1074164754,
    SS_IOCTL_TIOCSETC	= -2147060719,
    SS_IOCTL_TIOCLBIC	= -2147191682,
    SS_IOCTL_TIOCLBIS	= -2147191681,
    SS_IOCTL_TIOCLGET	= 0x4004747c,
    SS_IOCTL_TIOCLSET	= -2147191683,
};

/* struct definition (from simple-scalar/PISA) ************************/
/**********************************************************************/
struct ss_statbuf {
	uint032_t ss_st_dev;
	uint032_t ss_st_pad1[3];
	uint032_t ss_st_ino;
	uint032_t ss_st_mode;
	uint032_t ss_st_nlink;
	uint032_t ss_st_uid;
	uint032_t ss_st_gid;
	uint032_t ss_st_rdev;
	uint032_t ss_st_pad2[2];
	uint032_t ss_st_size;
	uint032_t ss_st_pad3;
	uint032_t ss_st_atime;
	uint032_t ss_st_atime_nsec;
	uint032_t ss_st_mtime;
	uint032_t ss_st_mtime_nsec;
	uint032_t ss_st_ctime;
	uint032_t ss_st_ctime_nsec;
	uint032_t ss_st_blksize;
	uint032_t ss_st_blocks;
	uint032_t ss_st_pad4[14];
};

struct ss_sgttyb {
    int008_t sg_ispeed;     /* input speed */
    int008_t sg_ospeed;     /* output speed */
    int008_t sg_erase;      /* erase character */
    int008_t sg_kill;       /* kill character */
    int032_t sg_flags;      /* mode flags */
};

struct ss_timeval {
    int032_t ss_tv_sec;		/* seconds */
    int032_t ss_tv_usec;		/* microseconds */
};

struct ss_timezone
{
    int032_t ss_tz_minuteswest;/* minutes west of Greenwich */
    int032_t ss_tz_dsttime;	/* type of dst correction */
};

struct ss_tms {
    uint032_t ss_tms_utime;
    uint032_t ss_tms_stime;
    uint032_t ss_tms_cutime;
    uint032_t ss_tms_cstime;
};

struct ss_rusage {
    struct ss_timeval ss_ru_utime;
    struct ss_timeval ss_ru_stime;
    int032_t ss_ru_maxrss;
    int032_t ss_ru_ixrss;
    int032_t ss_ru_idrss;
    int032_t ss_ru_isrss;
    int032_t ss_ru_minflt;
    int032_t ss_ru_majflt;
    int032_t ss_ru_nswap;
    int032_t ss_ru_inblock;
    int032_t ss_ru_oublock;
    int032_t ss_ru_msgsnd;
    int032_t ss_ru_msgrcv;
    int032_t ss_ru_nsignals;
    int032_t ss_ru_nvcsw;
    int032_t ss_ru_nivcsw;
};

struct ss_rlimit
{
    int ss_rlim_cur;		/* current (soft) limit */
    int ss_rlim_max;		/* maximum value for rlim_cur */
};

struct ss_dirent
{
    uint032_t d_ino;
    uint032_t d_off;
    uint016_t d_reclen;
    uint008_t d_type;
    char d_name[256];		/* We must not include limits.h! */
};

struct linux_dirent {
    long           d_ino;
    off_t          d_off;
    unsigned short d_reclen;
    char           d_name[];
};

/* initialize static members ******************************************/
/**********************************************************************/
uint064_t Mips::timeclk = 0;
uint064_t Mips::timesclk = 0;
uint064_t Mips::tofdayclk = 0;
time_t Mips::inittime = 0;
struct tms Mips::inittms;
struct timeval Mips::inittimeval;
struct timezone Mips::inittimezone;

/**********************************************************************/
void Mips::syscall()
{
    int a0 = (int)as->r[REG_A0];
    int a1 = (int)as->r[REG_A1];
    int a2 = (int)as->r[REG_A2];
    int a3 = (int)as->r[REG_A3];
#ifdef STRACE
    Printf deb_printf = printf;
#else
    Printf deb_printf = silentprintf;
#endif
    MainMemory *mem = (MainMemory *)(board->mmap->dev);

    switch (as->r[REG_V0]) {
    case SYS_EXIT:
        deb_printf("exit()\n");
        state = CPU_STOP;
        break;
    case SYS_FORK: { // broken ?
        as->r[REG_V0] = fork();
        deb_printf("fork()=%d\n", as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) {
            as->r[REG_A3] = 0;
            if(as->r[REG_V0] != 0)
                as->r[3] = 0;
        } else {
            fetal("## SYS_FORK failed.\n");
        }
        break;
    }
//     case SYS_VFORK: { // broken ?
//         as->r[REG_V0] = vfork();
//         deb_printf("vfork()=%d\n", as->r[REG_V0]);
//         if((int)as->r[REG_V0] != -1) 
//             as->r[REG_A3] = 0;
//         else
//             fetal("## SYS_VFORK failed.\n");
//         if((int)as->r[REG_V0] != 0) {
//             as->r[3] = 0;
//             as->r[REG_A3] = 0;
//         } 
//         break;
//     }
    case SYS_READ: {
        uint008_t tmp[a2];
        as->r[REG_V0] = read(a0, tmp, a2);
        deb_printf("read(%d, 0x%x, %d)=%d\n", a0, a1, a2, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        // write back result to simulator's memory
        mem->writenb(a1, as->r[REG_V0], tmp);
        break;
    }
    case SYS_WRITE: {
        uint008_t tmp[a2];
        mem->readnb(a1, a2, tmp);
        if(board->silent) { // silent mode
            int fd2 = open("/dev/null", O_CREAT | O_WRONLY, S_IRWXU);
            as->r[REG_V0] = write(fd2, tmp, a2);
            close(fd2);
        } else {
            as->r[REG_V0] = write(a0, tmp, a2);
        }        
        // error check
        if((int)as->r[REG_V0] == a2)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("write(%d,0x%x,%d)=%d(%d)\n",
                   a0, a1, a2, (int)as->r[REG_V0], (int)as->r[REG_A3]);
        break;
    }
    case SYS_WRITEV: {
        int i;
        char *buf;
        struct iovec iov[a2];
        
        mem->readnb(a1, a2*sizeof(struct iovec), (uint008_t *)iov);
        
        for (i=0; i < a2; i++) {
            iov[i].iov_base = (char *)iov[i].iov_base;
            iov[i].iov_len = iov[i].iov_len;
            if (iov[i].iov_base != NULL) {
                buf = (char *)calloc(iov[i].iov_len, sizeof(char));
                if (!buf)
                    fetal("out of virtual memory in SYS_writev");
                mem->readnb((uint032_t)(uint064_t)iov[i].iov_base,
                            iov[i].iov_len, (uint008_t *)buf);
                iov[i].iov_base = buf;
            }
        }
        as->r[REG_V0] = writev(a0, iov, a2);
        
        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        
        /* free all the allocated memory */
        for (i=0; i < a2; i++) {
            if (iov[i].iov_base) {
                free(iov[i].iov_base);
                iov[i].iov_base = NULL;
            }
        }
        break;
    }
    case SYS_OPEN: {
        uint008_t buf[SYSCALL_BUFSIZE];
        int flags_ia32 = 0;
        // convert flag
        if(a1 & MIPS_O_RDONLY)   flags_ia32 |= O_RDONLY;   
        if(a1 & MIPS_O_WRONLY)   flags_ia32 |= O_WRONLY;   
        if(a1 & MIPS_O_RDWR)     flags_ia32 |= O_RDWR;     
        if((a1 & MIPS_O_RDWR) && (a1 & MIPS_O_WRONLY))
            flags_ia32 |= O_ACCMODE;  
        if(a1 & MIPS_O_APPEND)   flags_ia32 |= O_APPEND;   
#if !defined (CYGWIN) && !defined(FreeBSD)
        if(a1 & MIPS_O_ASYNC)    flags_ia32 |= O_ASYNC;
        if(a1 & MIPS_O_DIRECTORY)flags_ia32 |= O_DIRECTORY;    
#endif
        if(a1 & MIPS_O_CREAT)    flags_ia32 |= O_CREAT;    
        if(a1 & MIPS_O_EXCL)     flags_ia32 |= O_EXCL;     
        if(a1 & MIPS_O_NOCTTY)   flags_ia32 |= O_NOCTTY;   
        if(a1 & MIPS_O_NONBLOCK) flags_ia32 |= O_NONBLOCK;
        if(a1 & MIPS_O_SYNC)     flags_ia32 |= O_SYNC;     
        if(a1 & MIPS_O_TRUNC)    flags_ia32 |= O_TRUNC;
        int unknownflag = 
            a1 & ~(MIPS_O_RDONLY   | MIPS_O_WRONLY | MIPS_O_RDWR   |
                   MIPS_O_ACCMODE  | MIPS_O_APPEND | MIPS_O_ASYNC  |
                   MIPS_O_CREAT    | MIPS_O_EXCL   | MIPS_O_NOCTTY |
                   MIPS_O_NONBLOCK | MIPS_O_SYNC   | MIPS_O_TRUNC  |
                   MIPS_O_DIRECTORY);
        if(unknownflag) {
            fprintf(stderr, "## SYS_OPEN cannot decode flag(%x).\n",
                    unknownflag);
            exit(1);
        }

        memstrcpy(mem, a0, buf);
        as->r[REG_V0] = open((char*)buf, flags_ia32, a2);
        deb_printf("open(0x%x, %x, %x)=%d\n", a0, a1, a2, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_CLOSE: {
        if(a0 == 0 || a0 == 1 || a0 ==2) {
            as->r[REG_A3] = 0;
            deb_printf("close(%d)\n", a0);
            break;
        } else {
            as->r[REG_V0] = close(a0);
            deb_printf("close(%d)=%d\n", a0, as->r[REG_V0]);

            // error check
            if((int)as->r[REG_V0] != -1) 
                as->r[REG_A3] = 0;
            else {
                as->r[REG_V0] = errno;
                as->r[REG_A3] = 1;
            }
        }
        break;
    }
    case SYS_CREATE: {
        uint008_t buf[SYSCALL_BUFSIZE];
        memstrcpy(mem, a0, buf);
        as->r[REG_V0] = creat((const char *)buf, a1);
        deb_printf("create(%d, %d)=%d\n", a0, a1, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_UNLINK: {
        uint008_t buf[SYSCALL_BUFSIZE];
        memstrcpy(mem, a0, buf);
        as->r[REG_V0] = unlink((const char *)buf);
        deb_printf("unlink(%d)=%d\n", a0, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;        
    }
    case SYS_CHDIR: {
        uint008_t buf[SYSCALL_BUFSIZE];
        memstrcpy(mem, a0, buf);
        as->r[REG_V0] = chdir((const char *)buf);
        deb_printf("chdir(%d)=%d\n", a0, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_CHMOD: {
        uint008_t buf[SYSCALL_BUFSIZE];
        memstrcpy(mem, a0, buf);
        as->r[REG_V0] = chmod((const char *)buf, a1);
        deb_printf("chmod(%d,%d)=%d\n", a0, a1, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_CHOWN: {
        uint008_t buf[SYSCALL_BUFSIZE];
        memstrcpy(mem, a0, buf);
        as->r[REG_V0] = chown((const char *)buf, a1, a2);
        deb_printf("chown(%d,%d,%d)=%d\n", a0, a1, a2, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_BRK: {
        uint032_t addr = ROUND_UP(a0); // round page boudary

        // error check
        if(addr == 0) {
            as->r[REG_V0] = as->heapptr; // fixed: by yoshito
        } else if(/*addr >= as->heapptr && */addr < as->r[REG_SP]) {
            as->r[REG_V0] = 0;
            as->r[REG_A3] = 0;
            as->heapptr = addr;
        } else {
            as->r[REG_V0] = ENOMEM;
            as->r[REG_A3] = 1;
            fprintf(stderr, "## SYS_BRK there is no memory(%x)\n",a0);
        }
        deb_printf("brk(0x%x)=0x%x\n", a0, as->r[REG_V0]);
        break;
    }
    case SYS_LSEEK: {
        as->r[REG_V0] = (uint032_t)lseek(a0, (off_t)a1, a2);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("lseek(%d, 0x%x, %d)=0x%x\n", a0, a1, a2, as->r[REG_V0]);
        break;
    }
    case SYS_GETPID: {
        as->r[REG_V0] = getpid();

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("getpid()=%d\n", as->r[REG_V0]);
        break;
    }
    case SYS_GETUID: {
        as->r[REG_V0] = getuid();
        as->r[3] = geteuid();

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("getuid()=%d\n", as->r[REG_V0]);
        break;
    }
    case SYS_GETEUID: {
        as->r[REG_V0] = geteuid();

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("geteuid()=%d\n", as->r[REG_V0]);
        break;
    }
    case SYS_GETEGID: {
        as->r[REG_V0] = getegid();

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("geteuid()=%d\n", as->r[REG_V0]);
        break;
    }
    case SYS_ACCESS: {
        uint008_t buf[SYSCALL_BUFSIZE];
        memstrcpy(mem, a0, buf);
        as->r[REG_V0] = access((const char *)buf, a1);
        deb_printf("access(%d,%d)=%d\n", a0, a1, as->r[REG_V0]);

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_STAT: 
    case SYS_LSTAT: {
        uint008_t buf[SYSCALL_BUFSIZE];
        struct ss_statbuf ss_sbuf;
        struct stat sbuf;
        memstrcpy(mem, a0, buf);
        if(as->r[REG_V0] == SYS_STAT) {
            as->r[REG_V0] = stat((const char *)buf, &sbuf);
            deb_printf("stat(%d,%d)=%d\n", a0, a1, as->r[REG_V0]);
        } else if(as->r[REG_V0] == SYS_LSTAT) {
            as->r[REG_V0] = lstat((const char *)buf, &sbuf);
            deb_printf("lstat(%d,%d)=%d\n", a0, a1, as->r[REG_V0]);
        }

        // error check
        if((int)as->r[REG_V0] != -1) 
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        
        // convert stat
        ss_sbuf.ss_st_dev = sbuf.st_dev;
        // ss_sbuf.ss_pad = 0;
        ss_sbuf.ss_st_ino = sbuf.st_ino;
        ss_sbuf.ss_st_mode = sbuf.st_mode;
        ss_sbuf.ss_st_nlink = sbuf.st_nlink;
        ss_sbuf.ss_st_uid = sbuf.st_uid;
        ss_sbuf.ss_st_gid = sbuf.st_gid;
        ss_sbuf.ss_st_rdev = sbuf.st_rdev;
        // ss_sbuf.ss_pad1 = 0;
        ss_sbuf.ss_st_size = sbuf.st_size;
        ss_sbuf.ss_st_atime = sbuf.st_atime;
        // ss_sbuf.ss_st_spare1 = 0;
        ss_sbuf.ss_st_mtime = sbuf.st_mtime;
        // ss_sbuf.ss_st_spare2 = 0;
        ss_sbuf.ss_st_ctime = sbuf.st_ctime;
        // ss_sbuf.ss_st_spare3 = 0;
        ss_sbuf.ss_st_blksize = sbuf.st_blksize;
        ss_sbuf.ss_st_blocks = sbuf.st_blocks;
        // ss_sbuf.ss_st_gennum = 0;
        // ss_sbuf.ss_st_spare4 = 0;
        
        // writeback result to simulator's memory
        mem->writenb(a1, sizeof(struct ss_statbuf), (uint008_t *)&ss_sbuf);
        break;
    }
    case SYS_DUP: {
        as->r[REG_V0] = dup(a0);
        deb_printf("dup(%d)=%d\n",a0,as->r[REG_V0]);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_PIPE: {
        int fd[2];
        mem->readnb(a0, sizeof(fd), (uint008_t *)fd);
        as->r[REG_A3] = pipe(fd);
        as->r[REG_V0] = fd[0];
        as->r[3] = fd[1];
        deb_printf("pipe(%d)=%d\n",a0,as->r[REG_V0]);

        // error check
        if ((int)as->r[REG_V0] == -1) {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_GETGID : {
        as->r[REG_V0] = getgid();
        as->r[3] = getegid();
        deb_printf("getgid()=%d\n",as->r[REG_V0]);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_IOCTL: {
        char buf[NUM_IOCTL_BYTES];
        int local_req = 0;

        switch (a2) {
#ifdef TIOCGETP
        case SS_IOCTL_TIOCGETP:
            local_req = TIOCGETP;
            break;
#endif
#ifdef TIOCSETP
        case SS_IOCTL_TIOCSETP:
            local_req = TIOCSETP;
            break;
#endif
#ifdef TIOCGETP
        case SS_IOCTL_TCGETP:
            local_req = TIOCGETP;
            break;
#endif
#ifdef TCGETA
        case SS_IOCTL_TCGETA:
            local_req = TCGETA;
            break;
#endif
#ifdef TIOCGLTC
        case SS_IOCTL_TIOCGLTC:
            local_req = TIOCGLTC;
            break;
#endif
#ifdef TIOCSLTC
        case SS_IOCTL_TIOCSLTC:
            local_req = TIOCSLTC;
            break;
#endif
#ifdef TIOCGWINSZ
        case SS_IOCTL_TIOCGWINSZ:
            local_req = TIOCGWINSZ;
            break;
#endif
#ifdef TCSETAW
        case SS_IOCTL_TCSETAW:
            local_req = TCSETAW;
            break;
#endif
#ifdef TIOCGETC
        case SS_IOCTL_TIOCGETC:
            local_req = TIOCGETC;
            break;
#endif
#ifdef TIOCSETC
        case SS_IOCTL_TIOCSETC:
            local_req = TIOCSETC;
            break;
#endif
#ifdef TIOCLBIC
        case SS_IOCTL_TIOCLBIC:
            local_req = TIOCLBIC;
            break;
#endif
#ifdef TIOCLBIS
        case SS_IOCTL_TIOCLBIS:
            local_req = TIOCLBIS;
            break;
#endif
#ifdef TIOCLGET
        case SS_IOCTL_TIOCLGET:
            local_req = TIOCLGET;
            break;
#endif
#ifdef TIOCLSET
        case SS_IOCTL_TIOCLSET:
            local_req = TIOCLSET;
            break;
#endif
        }
        
#if !defined(TIOCGETP) && (defined(linux) || defined(__CYGWIN32__))
        if (!local_req && a1 == SS_IOCTL_TIOCGETP)  {
            struct termios lbuf;
            struct ss_sgttyb buf;
            as->r[REG_V0] = tcgetattr(a0, &lbuf);

            /* translate results */
            buf.sg_ispeed = lbuf.c_ispeed;
            buf.sg_ospeed = lbuf.c_ospeed;
            buf.sg_erase = lbuf.c_cc[VERASE];
            buf.sg_kill = lbuf.c_cc[VKILL];
            buf.sg_flags = 0;   /* FIXME: this is wrong... */

            mem->writenb(a2, sizeof(struct ss_sgttyb), (uint008_t *)&buf);
            
            // error check
            if ((int)as->r[REG_V0] != -1)
                as->r[REG_A3] = 0;
            else {
                as->r[REG_V0] = errno;
                as->r[REG_A3] = 1;
            }
        } else
#endif
        if (!local_req)  {
            as->r[REG_V0] = 0;
            as->r[REG_A3] = 0;
        } else {
            if (a2 != 0)
                mem->readnb(a2, NUM_IOCTL_BYTES, (uint008_t *)buf);
            as->r[REG_V0] = ioctl(a0, local_req, buf);
            if (a2 != 0)
                mem->writenb(a2, NUM_IOCTL_BYTES, (uint008_t *)buf);

            // error check
            if ((int)as->r[REG_V0] != -1)
                as->r[REG_A3] = 0;
            else {
                as->r[REG_V0] = errno;
                as->r[REG_A3] = 1;
            }
        }
        deb_printf("ioctrl(%d, %d, %d)=%d\n", a0, a1, a2, as->r[REG_V0]);
        break;
    }
    case SYS_FSTAT: {
        struct ss_statbuf ss_sbuf;
        struct stat sbuf;
        as->r[REG_V0] = fstat(a0, &sbuf);
        deb_printf("fstat(%d, 0x%x)=%d\n", a0, a1, as->r[REG_V0]);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }

        // convert sbuf format
        ss_sbuf.ss_st_dev = sbuf.st_dev;
        // ss_sbuf.ss_pad = 0;
        ss_sbuf.ss_st_ino = sbuf.st_ino;
        ss_sbuf.ss_st_mode = sbuf.st_mode;
        ss_sbuf.ss_st_nlink = sbuf.st_nlink;
        ss_sbuf.ss_st_uid = sbuf.st_uid;
        ss_sbuf.ss_st_gid = sbuf.st_gid;
        ss_sbuf.ss_st_rdev = sbuf.st_rdev;
        // ss_sbuf.ss_pad1 = 0;
        ss_sbuf.ss_st_size = sbuf.st_size;
        ss_sbuf.ss_st_atime = sbuf.st_atime;
        // ss_sbuf.ss_st_spare1 = 0;
        ss_sbuf.ss_st_mtime = sbuf.st_mtime;
        // ss_sbuf.ss_st_spare2 = 0;
        ss_sbuf.ss_st_ctime = sbuf.st_ctime;
        // ss_sbuf.ss_st_spare3 = 0;
        ss_sbuf.ss_st_blksize = sbuf.st_blksize;
        ss_sbuf.ss_st_blocks = sbuf.st_blocks;
        // ss_sbuf.ss_st_gennum = 0;
        // ss_sbuf.ss_st_spare4 = 0;
        
        // writeback result to simulator's memory
        mem->writenb(a1, sizeof(struct ss_statbuf), (uint008_t *)&ss_sbuf);
        break;
    }
    case SYS_DUP2: {
        as->r[REG_V0] = dup2(a0,a1);
        deb_printf("dup2(%d,%d)=%d\n",a0,a1,as->r[REG_V0]);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_FCNTL: {
        as->r[REG_V0] = fcntl(a0,a1,a2);
        deb_printf("fcntl(%d,%d,%d)=%d\n",a0,a1,a2,as->r[REG_V0]);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        break;
    }
    case SYS_MMAP: {
        int a4, a5;
        mem->read4b(as->r[REG_SP] +  16, (uint032_t*)&a4);
        mem->read4b(as->r[REG_SP] +  20, (uint032_t*)&a5);

        // decide mapping address
        uint032_t addr;
        if(a0 == 0) {
            // select map addr by kernel
            addr = ROUND_UP(as->heapptr);
            while(mem->pagetable[addr/PAGE_SIZE] != NULL) {
                addr += PAGE_SIZE;
                if( addr >= MEM_SIZE_DEF ) 
                    fetal("## SYS MMAP can't create mapping.\n");
            }
        } else {
            // map addr <= a0
            addr = ROUND_UP(a0);
        }

        // create mapping
        if(a3 & MIPS_MAP_ANONYMOUS) {
//             uint008_t buf[a1];
//             //memset(buf, 0, sizeof(uint008_t)*a1);
//             if(a3 & MIPS_MAP_GROWSDOWN) 
//                 mem->writenb((uint032_t)addr-a1, a1, buf);
//             else
//                 mem->writenb((uint032_t)addr, a1, buf);
            as->heapptr = addr + a1;
        } else {
            uint032_t readcnt;
            uint008_t buf[a1 + a5];
            readcnt = read(a4, buf, a1 + a5);
            
            mem->writenb((uint032_t)addr, a1, &buf[a5]);
            as->heapptr = addr + readcnt;
        }
        if(a3 & MIPS_MAP_GROWSDOWN)
            fetal("## SYS_MMAP MAP_GROWSDOWN is not implemented.\n");

        // error check
        if(addr > as->r[REG_SP]) {
            as->r[REG_V0] = ENOMEM;
            as->r[REG_A3] = 1;
            fprintf(stderr, "## SYS_MMAP there is no memory(%x)\n",a0);
        } else {
            as->r[REG_V0] = addr;
            as->r[REG_A3] = 0;
        }
        deb_printf("mmap(0x%x, 0x%x, %d, 0x%x, %d, 0x%x)=0x%x\n",
                   a0, a1, a2, a3, a4, a5, as->r[REG_V0]);
        break;
    }
    case SYS_GETRLIMIT: 
    case SYS_SETRLIMIT: {
        struct ss_rlimit ss_rl;
        struct rlimit rl;
            
        mem->readnb(a1, sizeof(struct ss_rlimit), (uint008_t *)&ss_rl);
        // convert
        rl.rlim_cur = ss_rl.ss_rlim_cur;
        rl.rlim_max = ss_rl.ss_rlim_max;            

        if (as->r[REG_V0] == SYS_GETRLIMIT)
            as->r[REG_V0] = getrlimit(a0, &rl);
        else 
            as->r[REG_V0] = setrlimit(a0, &rl);
            
        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        // convert
        ss_rl.ss_rlim_cur = rl.rlim_cur;
        ss_rl.ss_rlim_max = rl.rlim_max;

        mem->writenb(a1, sizeof(struct ss_rlimit), (uint008_t *)&ss_rl);

        if (as->r[REG_V0] == SYS_GETRLIMIT)
            deb_printf("getrlimit(%d, %d)=%d\n", a0, a1, as->r[REG_V0]);
        else 
            deb_printf("setrlimit(%d, %d)=%d\n", a0, a1, as->r[REG_V0]);
        break;
    }
    case SYS_TIMES: {
        struct tms buf;
        struct ss_tms ssbuf;
        as->r[REG_V0] = (uint032_t)times(&buf);

        if(timesclk == 0) { // first call of sys_times
            // convert format
            ssbuf.ss_tms_utime  = (uint032_t)buf.tms_utime;
            ssbuf.ss_tms_stime  = (uint032_t)buf.tms_stime;
            ssbuf.ss_tms_cutime = (uint032_t)buf.tms_cutime;
            ssbuf.ss_tms_cstime = (uint032_t)buf.tms_cstime;
            inittms = buf;
        } else {
            uint064_t inc =
                (chip->cycle - timesclk)*sysconf(_SC_CLK_TCK)/FREQUENCY;
            ssbuf.ss_tms_utime  = (uint032_t)(inittms.tms_utime + inc);
            ssbuf.ss_tms_stime  = (uint032_t)(inittms.tms_stime + inc);
            ssbuf.ss_tms_cutime = (uint032_t)(inittms.tms_cutime + inc);
            ssbuf.ss_tms_cstime = (uint032_t)(inittms.tms_cstime + inc);
        }

        // write result to memory
        if(a0)
            mem->writenb(a0, sizeof(struct ss_tms), (uint008_t *)&ssbuf);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("times(%d)=%d\n", a0, as->r[REG_V0]);
        break;
    }
    case SYS_TIME: {
        time_t buf;
        uint032_t ssbuf;
        uint064_t ret = time(&buf);

        if(timeclk == 0) { // first call of sys_time
            // convert format
            ssbuf = (uint032_t)buf;
            timeclk = chip->cycle;
            inittime = buf;
            as->r[REG_V0] = (uint032_t)ret;
        } else {
            uint064_t inc = (chip->cycle - timeclk)/FREQUENCY;
            ssbuf = (uint032_t)(inittime + inc);
            as->r[REG_V0] = ssbuf;;
        }

        // write result to memory
        if(a0)
            mem->writenb(a0, sizeof(ssbuf), (uint008_t *)&ssbuf);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("time(%d)=%d\n", a0, as->r[REG_V0]);
        break;
    }
    case SYS_GETTIMEOFDAY: {
        struct timeval tv;
        struct timezone tz;
        struct ss_timeval sstv;
        struct ss_timezone sstz;
        as->r[REG_V0] = (uint032_t)gettimeofday(&tv, &tz);
        
        if(tofdayclk == 0) { // first call of sys_gettimeofday
            // convert format
            sstv.ss_tv_sec = (uint032_t)tv.tv_sec;
            sstv.ss_tv_usec = (uint032_t)tv.tv_usec;
            sstz.ss_tz_minuteswest = (uint032_t)tz.tz_minuteswest;
            sstz.ss_tz_dsttime = (uint032_t)tz.tz_dsttime;
            tofdayclk = chip->cycle;
            inittimeval = tv;
            inittimezone = tz;
        } else {
            uint064_t inc = (chip->cycle - tofdayclk)/FREQUENCY;
            uint064_t uinc = // is it correct?
                ((chip->cycle - tofdayclk)/(FREQUENCY/1000000)); 
            sstv.ss_tv_sec = (uint032_t)(inittimeval.tv_sec + inc);
            sstv.ss_tv_usec = 
                (uint032_t)((inittimeval.tv_usec + uinc) % 1000000);
            sstz.ss_tz_minuteswest = 
                (uint032_t)(inittimezone.tz_minuteswest + inc/60);
            sstz.ss_tz_dsttime = inittimezone.tz_dsttime; 
        }

        // write result to memory
        if(a0)
            mem->writenb(a0, sizeof(sstv), (uint008_t *)&sstv);
        if(a1)
            mem->writenb(a1, sizeof(sstz), (uint008_t *)&sstz);

        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("gettimeofday(0x%x, 0x%x)=%d\n", a0,a1,as->r[REG_V0]);
        break;
    }
    case SYS_GETRUSAGE: {
        struct rusage local_rusage;
        struct ss_rusage rusage;
        
        as->r[REG_V0] = getrusage(a0, &local_rusage);
        
        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        
        // convert from host rusage structure to target format
        rusage.ss_ru_utime.ss_tv_sec = local_rusage.ru_utime.tv_sec;
        rusage.ss_ru_utime.ss_tv_usec = local_rusage.ru_utime.tv_usec;
        rusage.ss_ru_utime.ss_tv_sec = local_rusage.ru_utime.tv_sec;
        rusage.ss_ru_utime.ss_tv_usec =local_rusage.ru_utime.tv_usec;
        rusage.ss_ru_stime.ss_tv_sec = local_rusage.ru_stime.tv_sec;
        rusage.ss_ru_stime.ss_tv_usec = local_rusage.ru_stime.tv_usec;
        rusage.ss_ru_stime.ss_tv_sec = local_rusage.ru_stime.tv_sec;
        rusage.ss_ru_stime.ss_tv_usec = local_rusage.ru_stime.tv_usec;
        rusage.ss_ru_maxrss = local_rusage.ru_maxrss;
        rusage.ss_ru_ixrss = local_rusage.ru_ixrss;
        rusage.ss_ru_idrss = local_rusage.ru_idrss;
        rusage.ss_ru_isrss = local_rusage.ru_isrss;
        rusage.ss_ru_minflt = local_rusage.ru_minflt;
        rusage.ss_ru_majflt = local_rusage.ru_majflt;
        rusage.ss_ru_nswap = local_rusage.ru_nswap;
        rusage.ss_ru_inblock = local_rusage.ru_inblock;
        rusage.ss_ru_oublock = local_rusage.ru_oublock;
        rusage.ss_ru_msgsnd = local_rusage.ru_msgsnd;
        rusage.ss_ru_msgrcv = local_rusage.ru_msgrcv;
        rusage.ss_ru_nsignals = local_rusage.ru_nsignals;
        rusage.ss_ru_nvcsw = local_rusage.ru_nvcsw;
        rusage.ss_ru_nivcsw = local_rusage.ru_nivcsw;
        
        // writeback result to simulator's memory
        mem->writenb(a1, sizeof(struct ss_rusage), (uint008_t *)&rusage);

        deb_printf("getrusage(%d, %d)=%d\n",a0, a1, as->r[REG_V0]);
        break;
    }
    case SYS_UTIME: {
        uint008_t buf[SYSCALL_BUFSIZE];
        memstrcpy(mem, a0, buf);
        
        if (a1 == 0) {
            if(board->silent)
                as->r[REG_V0] = 0;
            else
                as->r[REG_V0] = utime((const char*)buf, NULL);
        } else {
            struct timeval tval[2];
            struct ss_timeval ss_tval[2];
            mem->readnb(a1, 2*sizeof(struct ss_timeval), 
                        (uint008_t *)ss_tval);
            
            tval[0].tv_sec = ss_tval[0].ss_tv_sec;
            tval[0].tv_usec = ss_tval[0].ss_tv_usec;
            tval[1].tv_sec = ss_tval[1].ss_tv_sec;
            tval[1].tv_usec = ss_tval[1].ss_tv_usec;

            if(board->silent)
                as->r[REG_V0] = 0;
            else    
                as->r[REG_V0]= utimes((const char*)buf, tval);
        }
        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }

        deb_printf("utime(%d, %d)=%d\n",a0, a1, as->r[REG_V0]);
        break;
    }
    case SYS_MUNMAP: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        deb_printf("munmap(%x, %d)=%d\n", a0, a1, as->r[REG_V0]);
        break;
    }
    case SYS_MREMAP: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_MREMAP not implemented\n");
        break;
    }
    case SYS_GETDENTS: { // bug?
#if defined (MACOSX) || defined(CYGWIN)
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_GETDENTS not implemented\n");
#else
        char buf[a2];
        struct linux_dirent *d;
        struct ss_dirent ss_d;

        as->r[REG_V0] = ::syscall(SYS_getdents, a0, buf, a2);

        for (uint i = 0; i < as->r[REG_V0];) {
            d = (struct linux_dirent *) (buf + i);
            ss_d.d_ino = d->d_ino;
            ss_d.d_off = d->d_off;
            ss_d.d_reclen = d->d_reclen;
            ss_d.d_type = *(buf + i + d->d_reclen - 1);
            strcpy(ss_d.d_name, (const char *)d->d_name);
            mem->writenb(a1+i, sizeof(struct ss_dirent),
                         (uint008_t *)&ss_d);
            mem->write1b(a1+i+d->d_reclen-1,
                         (uint008_t)*(buf + i + d->d_reclen - 1));
            i += d->d_reclen;
        }
        // error check
        if ((int)as->r[REG_V0] != -1)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("getdents(%d,%d,%d)=%d\n",a0, a1, a2, as->r[REG_V0]);
#endif
        break;
    }
    case SYS_GETCWD: {
        char buf[a1];
        uint064_t ret = (uint064_t)getcwd(buf, a1);
        as->r[REG_V0] = a0;
        mem->writenb(a0, strlen(buf)+1, (uint008_t *)&buf);
        if (ret != 0)
            as->r[REG_A3] = 0;
        else {
            as->r[REG_V0] = errno;
            as->r[REG_A3] = 1;
        }
        deb_printf("getcwd(0x%x, %d)=0x%x\n", a0, a1, as->r[REG_V0]);
        break;
    }
    case SYS_RT_SIGPROCMASK: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_RT_SIGPROCMASK not implemented\n");
        break;
    }
    case SYS_FTRUNCATE: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_FTRUNCATE not implemented\n");
        break;
    }
    case SYS_KILL: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_KILL not implemented\n");
        break;
    }
    case SYS_RMDIR: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_RMDIR not implemented\n");
        break;
    }
    case SYS_RT_SIGACTION: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_RT_SIGACTION not implemented\n");
        break;
    }
    case SYS_FUTEX: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_FUTEX not implemented\n");
        break;
    }
    case SYS_UNAME: {
        as->r[REG_V0] = as->r[REG_A3] = 0;
        fprintf(stderr, "## SYS_UNAME not implemented\n");
        break;
    }
    default:
        printf("!! unknown syscall #%d (see asm/unistd.h). Skip.\n",
               as->r[REG_V0]);
        as->r[REG_V0] = as->r[REG_A3] = 0;
        break;
    }
}

/**********************************************************************/
