/**********************************************************************/
/* Many-Core Architecture Research Project       Arch Lab. TOKYO TECH */
/**********************************************************************/
#include "network.h"

/**********************************************************************/
/**
 * Inbuf constructor
 */
Inbuf::Inbuf() {
    _head = 0;
    _tail = 0;
    _num_entry = 0;
}

/**********************************************************************/
/**
 * get the number of flits saved in the buffer
 * \return the number of flits
 */
int Inbuf::num_entry()
{
    return _num_entry;
}

/**********************************************************************/
/**
 * check whether the buffer has one or more flits or not
 * \return 0: not empty, 1: empty
 */
bool Inbuf::isempty()
{
    return (_num_entry == 0);
}

/**********************************************************************/
/**
 * check whether the buffer has full flits or not
 * \return 0: not full, 1: full
 */
bool Inbuf::isfull()
{
    return (_num_entry == BUF_SIZE);
}

/**********************************************************************/
/**
 * enqueue a flit into the buffer
 * \param data flit to be enqueued
 * \return true: succeed, false: failed
 */
bool Inbuf::enq(flit_t data)
{
    if (isfull()) {
        fprintf(stderr, "## Enqueue is failed. Inbuf is full.\n");
        return false;
    }
    else {
        buf[_tail] = data;
        _tail = (_tail == BUF_SIZE - 1) ? 0 : _tail + 1;
        _num_entry++;
        return true;
    }
}

/**********************************************************************/
/**
 * dequeue a flit from the buffer
 * \return true: succeed, false: failed
 */
bool Inbuf::deq()
{
    if (isempty()) {
        fprintf(stderr, "## Dequeue is failed. Inbuf is empty.\n");
        printf("## Dequeue is failed. Inbuf is empty.\n");
        printf("### this:%p. num_entry:%d\n", this, _num_entry);
        return false;
    }
    else {
        _head = (_head == BUF_SIZE - 1) ? 0: _head + 1;
        _num_entry--;
        return true;
    }
}

/**********************************************************************/
/**
 * get the flit saved at the head of the buffer
 * \return flit_t*
 */
flit_t *Inbuf::head()
{
    if (isempty()) return NULL;
    else           return &buf[_head];
}

/**********************************************************************/
/**
 * routing computation
 */
void Inbuf::rtcomp(uint032_t id) {
    if(isempty() || !(head()->type & FLIT_HEADER))
        request = 0;
    else {
        uint016_t tmp = head()->header.coreid;
        if(IDX(id) > IDX(tmp))      request = 1 << W;
        else if(IDX(id) < IDX(tmp)) request = 1 << E;
        else if(IDY(id) < IDY(tmp)) request = 1 << S;
        else if(IDY(id) > IDY(tmp)) request = 1 << N;
        else                        request = 1 << C;
    }
}
