#include "network.h"

/**********************************************************************/
/**
 * Incc32 constructor
 */
Incc32::Incc32() {
    core_incc_if = NULL;
    dma_send_state = INIT;
    dma_type = DMA_NA;
}

/**********************************************************************/
/**
 * Incc32 destructor
 */
Incc32::~Incc32() {
    mem = NULL;
//    delete getreqbuf;
//    delete nackmap;
}

/**********************************************************************/
/**
 * initialize Incc32
 * /param env Env *
 * /param id the node id of the INCC32
 */
void Incc32::init(Env *env, int id) {
    int entry = env->id2entry(id);
    this->env = env;
    
    mem = (MainMemory*)env->node[entry].mem;
    mem->write4b(0, 0);
    memmap = mem->getrealaddr(0);
    // initialize memory map
    *(int*)((char*)memmap + SC_MMAP_NODE_ID) = id;
    *(int*)((char*)memmap + SC_MMAP_NOC_SIZE) = env->nocrank;
    *(int*)((char*)memmap + SC_MMAP_TIMER) = 0;
    *(int*)((char*)memmap + SC_MMAP_TIMER_D) = 0;
    *(int*)((char*)memmap + SC_MMAP_INST_LOOP) = 0x1000ffff;
    *(int*)((char*)memmap + SC_MMAP_INST_LOOP__) = 0x00000000;
    *(int*)((char*)memmap + SC_MMAP_PC) = 0;
    // for server node
    *(int*)((char*)memmap + SC_MMAP_NODE_SERVER) = 0;
    // for incc write addr
    *(int*)((char*)memmap + SC_MMAP_INCC_WRIGHT_ADDR) = 0;

    // get request buffer
    getreqbuf = new Cmdbuf(GETREQBUF_SIZE);
    
    // allocate bitmap 
    nackmap = new Bitmap((env->nr_node + 
                          sizeof(unsigned int) - 1)/sizeof(unsigned int));
    
    core_incc_if = (commcmd_t*)
        ((char*)memmap + SC_MMAP_DMA_PUT_REMOTE_ID) - 1;
    
    for(int i = 1; i < DMA_NUM_CMD; i++)
        core_incc_if[i].cmd = 0;
    
    coreid = (int*)((char*)memmap + SC_MMAP_NODE_ID);
    dma_type = DMA_NA;
    dma_send_state = INIT;
    ordy = 1;
    router = env->node[env->id2entry(id)].router;
}

/**********************************************************************/
/**
 * update all registers in the INCC32
 */
void Incc32::update() {
    // outbuf
    outbuf = _outbuf;
    switch(cmd_send.cmd) {
    case 1://DMA_PUT
        if(router->ordy_[C])
            outbuf.type |= FLIT_VALID;
        break;
    case 2://DMA_PUT64
        if(router->ordy_[C])
            outbuf.type |= FLIT_VALID;
        break;
    default:
        if(dma_type == DMA_NACK && router->ordy_[C])
            outbuf.type |= FLIT_VALID;
        break;
    }

    // pe_incc_if
    if(_dma_send_state == INIT && dma_type_ == DMA_PUT && router->ordy_[C])
        core_incc_if[DMA_PUT].cmd = DMA_NA;
    else if(_getreqbuf.cmd)
        core_incc_if[DMA_GET].cmd = DMA_NA;

    // send count
    if(dma_send_state_ != INIT && router->ordy_[C])
        send_count = ((_outbuf.type & FLIT_TAILER) ? 0 : send_count + 1);

    // send command register for dynamic router
    switch(dma_type_) {
    case DMA_NA:
        switch(_dma_type) {
        case DMA_NA:
            break;
        case DMA_PUT:
            cmd_send = core_incc_if[_dma_type];
            break;
        case DMA_GET:
            cmd_send = *(commcmd_t*)getreqbuf->head();
            if(router->ordy_[C])
                getreqbuf->deq();
            break;
        case DMA_NACK:
            nackentry = nackmap->lsb();
            nackmap->reset(nackentry);
            cmd_send.remote_id = env->entry2id(nackentry);
            break;
        }
        break;
    case DMA_PUT:
        switch(cmd_send.cmd) {
        case 1://DMA_PUT
            if(router->ordy_[C] && (_outbuf.type & FLIT_DATA)) {
                cmd_send.remote_addr += cmd_send.remote_stride;
                cmd_send.local_addr += cmd_send.local_stride;
                cmd_send.size -= 4;
            }
            break;
        case 2://DMA_PUT64
            if(router->ordy_[C] && (_outbuf.type & FLIT_DATA)) {
                *(uint064_t*)&cmd_send.remote_addr += 4;
                cmd_send.local_addr += cmd_send.local_stride;
                cmd_send.size -= 4;
            }
            break;
        }
        break;
    case DMA_GET:
        if(router->ordy_[C] && (_outbuf.type & FLIT_DATA)) {
            cmd_send.remote_addr += cmd_send.remote_stride;
            cmd_send.local_addr += cmd_send.local_stride;
            cmd_send.size -= 4;
        }
        break;
    case DMA_ACK:
        break;
    case DMA_NACK:
        break;
    }

    if(router->ordy_[C] &&_dma_type == DMA_NA)
        cmd_send.cmd = 0;

    // cmdbuf
    if(_getreqbuf.cmd) {
        if(getreqbuf->isfull())
            nackmap->set(env->id2entry(_getreqbuf.remote_id));
        else
            getreqbuf->enq((void*)&_getreqbuf);
    }
    
    // recv command register from dynamic router
    if(inbuf_.type & FLIT_VALID) {
        if(inbuf_.type & FLIT_HEADER)
            ;
        else if(inbuf_.type & FLIT_ADDR) {
            cmd_recv.local_addr = inbuf_.body.data;
            cmd_recv.local_stride = DEFAULT_STRIDE;
        }
        else if(inbuf_.type & FLIT_STRIDE)
            cmd_recv.local_stride = inbuf_.body.data;
        else if(inbuf_.type & FLIT_DATA) {
            if(cmd_recv.local_addr & 0x3) {
                fprintf(stderr, "### incc alignment error\n");
                exit(0);
            }
            mem->write4b(cmd_recv.local_addr, inbuf_.body.data);
            cmd_recv.local_addr += cmd_recv.local_stride;
        }
        // for memmap incc write addr
        *(int*)((char*)memmap + SC_MMAP_INCC_WRIGHT_ADDR) =
            cmd_recv.local_addr;
    }
    
    // dma_send_state
    if(dma_send_state_ == INIT || router->ordy_[C])
        dma_send_state = _dma_send_state;
    
    // dma_type
    if(router->ordy_[C])
        dma_type = _dma_type;
    
    // inbuf
    inbuf = in;
    
    // ready (flow control)
    ordy = 1;
    irdy = router->ordy_[C];

    // print 
    if (env->log_mode == 1 || env->log_mode == 2)
        print_packet();
    if(env->log_mode == 2)
        print_flit();
    if(env->incc_activity)
        plot();
}
    
/**********************************************************************/
/**
 * transfer flits between INCC32 and router,
 * calculate the next state
 */
void Incc32::comm() {
    const int lut[][4] = {
        {INIT},
        {INIT, PUT_FH, PUT64_FH},
        {INIT, GET_FH},
        {INIT, ACK_FH},
        {INIT, NACK_FH},
    };

    // in
    in = router->out[C];

    // rdy
    ordy_ = ordy;
    irdy_ = irdy;

    // inbuf
    inbuf_ = inbuf;

    // dma_send_state
    dma_send_state_ = dma_send_state;

    // _dma_type
    dma_type_ = dma_type;
    
    // _getreqbuf
    if(core_incc_if[DMA_GET].cmd != DMA_NA)
        _getreqbuf = core_incc_if[DMA_GET];
    else
        _getreqbuf.cmd = 0;
    
    // _dma_send_state, _dma_type, _outbuf
    switch(dma_send_state_) {
    case INIT:
        // select next dma cmd
        if(core_incc_if[DMA_PUT].cmd) {
            _dma_send_state = lut[DMA_PUT][core_incc_if[DMA_PUT].cmd];
            _dma_type = DMA_PUT;
        }
        else if(! getreqbuf->isempty()) {
            _dma_send_state = lut[DMA_GET]
                [((commcmd_t*)getreqbuf->head())->cmd];
            _dma_type = DMA_GET;
        }
        else if(! nackmap->isclear()) {
            _dma_send_state = lut[DMA_NACK][1];
            _dma_type = DMA_NACK;
        }
        else {
            _dma_send_state = INIT;
            _dma_type = DMA_NA;
        }            
        break;

    case PUT_FH:
        _outbuf.type = FLIT_HEADER;
        _outbuf.header.coreid = cmd_send.remote_id;
        _outbuf.header.tag = *coreid;
        _dma_send_state = PUT_FA;
        _dma_type = DMA_PUT;
        break;
    case PUT_FA:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = cmd_send.remote_addr;
        if(cmd_send.size == 0) {
            _dma_send_state = INIT;
            _dma_type = DMA_NA;
        }
        else if(cmd_send.remote_stride == 4) {
            _dma_send_state = PUT_FD;
            _dma_type = DMA_PUT;
        }
        else {
            _dma_send_state = PUT_FS;
            _dma_type = DMA_PUT;
        }
        break;
    case PUT_FS:
        _outbuf.type = FLIT_STRIDE;
        _outbuf.body.data = cmd_send.remote_stride;
        _dma_send_state = PUT_FD;
        _dma_type = DMA_PUT;
        break;
    case PUT_FD:
        if(cmd_send.local_addr & 0x3) {
            fprintf(stderr, "### incc alignment error\n");
            exit(0);
        }
        mem->read4b(cmd_send.local_addr, &_outbuf.body.data);
        if(cmd_send.size == 4) {
            _dma_send_state = INIT;
            _dma_type = DMA_NA;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else if(send_count == MAX_FLITS_PER_PACKET - 1) {
            _dma_send_state = PUT_FH;
            _dma_type = DMA_PUT;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else {
            _dma_send_state = PUT_FD;
            _dma_type = DMA_PUT;
            _outbuf.type = FLIT_DATA;
        }
        break;

    case GET_FH:
        _outbuf.type = FLIT_HEADER;
        _outbuf.header.coreid = cmd_send.remote_id;
        _outbuf.header.tag = 0;
        _dma_send_state = GET_FA;
        _dma_type = DMA_GET;
        break;
    case GET_FA:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = cmd_send.remote_addr;
        _dma_send_state = (cmd_send.remote_stride == 4) ?
            GET_FD : GET_FS;
        _dma_type = DMA_GET;
        break;
    case GET_FS:
        _outbuf.type = FLIT_STRIDE;
        _outbuf.body.data = cmd_send.remote_stride;
        _dma_send_state = GET_FD;
        _dma_type = DMA_GET;
        break;
    case GET_FD:
        if(cmd_send.local_addr & 0x3) {
            fprintf(stderr, "### incc alignment error\n");
            exit(0);
        }
        mem->read4b(cmd_send.local_addr, &_outbuf.body.data);
        if(cmd_send.size == 4) {
            _dma_send_state = ACK_FH;
            _dma_type = DMA_ACK;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else if(send_count == MAX_FLITS_PER_PACKET - 1) {
            _dma_send_state = GET_FH;
            _dma_type = DMA_GET;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else {
            _dma_send_state = GET_FD;
            _dma_type = DMA_GET;
            _outbuf.type = FLIT_DATA;
        }
        break;

    case ACK_FH:
        _outbuf.type = FLIT_HEADER;
        _outbuf.header.coreid = cmd_send.remote_id;
        _outbuf.header.tag = 0;
        _dma_send_state = ACK_FA;
        _dma_type = DMA_ACK;
        break;
    case ACK_FA:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = SC_MMAP_DMA_FLAG;
        _dma_send_state = ACK_FD;
        _dma_type = DMA_ACK;
        break;
    case ACK_FD:
        _outbuf.type = FLIT_DATA | FLIT_TAILER;
        _outbuf.body.data = DMAFLAG_ACK;
        _dma_send_state = INIT;
        _dma_type = DMA_NA;
        break;

    case NACK_FH:
        _outbuf.type = FLIT_HEADER;
        _outbuf.header.coreid = cmd_send.remote_id;
        _outbuf.header.tag = 0;
        _dma_send_state = NACK_FA;
        _dma_type = DMA_NACK;
        break;
    case NACK_FA:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = SC_MMAP_DMA_FLAG;
        _dma_send_state = NACK_FD;
        _dma_type = DMA_NACK;
        break;
    case NACK_FD:
        _outbuf.type = FLIT_DATA | FLIT_TAILER;
        _outbuf.body.data = DMAFLAG_NACK;
        _dma_send_state = INIT;
        _dma_type = DMA_NA;
        break;

    case PUT64_FH:
        _outbuf.type = FLIT_HEADER;
        _outbuf.header.coreid = cmd_send.remote_id;
        _outbuf.header.tag = 0;
        _dma_send_state = PUT64_FAL;
        _dma_type = DMA_PUT;
        break;
    case PUT64_FAL:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = cmd_send.remote_addr;
        _dma_send_state = PUT64_FAH;
        _dma_type = DMA_PUT;
        break;
    case PUT64_FAH:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = cmd_send.remote_stride;
        if(cmd_send.size == 0) {
            _dma_send_state = INIT;
            _dma_type = DMA_NA;
        }
        else {
            _dma_send_state = PUT64_FD;
            _dma_type = DMA_PUT;
        }
        break;
    case PUT64_FD:
        if(cmd_send.local_addr & 0x3) {
            fprintf(stderr, "### incc alignment error\n");
            exit(0);
        }
        mem->read4b(cmd_send.local_addr, &_outbuf.body.data);
        if(cmd_send.size == 4) {
            _dma_send_state = INIT;
            _dma_type = DMA_NA;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else if(send_count == MAX_FLITS_PER_PACKET - 1) {
            _dma_send_state = PUT64_FH;
            _dma_type = DMA_PUT;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else {
            _dma_send_state = PUT64_FD;
            _dma_type = DMA_PUT;
            _outbuf.type = FLIT_DATA;
        }
        break;

    }        
}

