#include "network.h"

/**********************************************************************/
/**
 * Incc64 constructor
 */
Incc64::Incc64() {
    ready_counter = 0;
}

/**********************************************************************/
/**
 * Incc64 destructor
 */
Incc64::~Incc64() {
    delete getreqbuf;
    delete nackmap;
}

/**********************************************************************/
/**
 * initialize INCC64
 * \param env Env *
 * \param the node id of the INCC64
 */
void Incc64::init(Env *env, int id) {
    int entry = env->id2entry(id);

    this->env = env;
    
    mem = (GMemory*)env->node[entry].mem;
    mem->write4b(0, 0);
    memmap = mem->getrealaddr((uint064_t)0);

    // initialize memory map
    *(int*)((char*)memmap + SC_MMAP_NODE_ID) = id;
    *(int*)((char*)memmap + SC_MMAP_NOC_SIZE) = env->nocrank;
    
    core_incc_if = (commcmd_t*)((char*)memmap +
                                SC_MMAP_DMA_PUT_REMOTE_ID) - 1;
    
//  initialize command registers
    for(int i = 1; i < DMA_NUM_CMD; i++)
        core_incc_if[i].cmd = DMA_NA;

//    getreqbuf = new Cmdbuf(GETREQBUF_SIZE);
    getreqbuf = new Cmdbuf(env->nr_node);
    
    // allocate bitmap 
    nackmap = new Bitmap((env->nr_node + 
                          sizeof(unsigned int) - 1)/sizeof(unsigned int));
    
    coreid = (int*)((char*)memmap + SC_MMAP_NODE_ID);
    dma_type_ = _dma_type = dma_type = DMA_NA;
    dma_send_state = INIT;
    ordy = 1;
    router = env->node[env->id2entry(id)].router;

    getdata = new Getdata(mem);
}

/**********************************************************************/
/**
 * update all registers
 */
void Incc64::update() {

#define DMA_GET_BUF 0

    // outbuf
    outbuf = _outbuf;
#if DMA_GET_BUF
    if(dma_send_state_ != INIT && router->ordy_[C])
#else
    if((dma_send_state_ != INIT && router->ordy_[C]) &&
        (dma_send_state_ != GET_FD || readable))
#endif
        outbuf.type |= FLIT_VALID;

    // send command register

#if DMA_GET_BUF
    if(dma_send_state_ != INIT && router->ordy_[C]) {
#else
    if((dma_send_state_ != INIT && router->ordy_[C]) && 
       (dma_send_state_ != GET_FD || readable)) {
#endif
        send_count = ((_outbuf.type & FLIT_TAILER) ? 0 : send_count + 1);
        if(_dma_send_state == INIT && _dma_type != DMA_NA)
            core_incc_if[_dma_type].cmd = DMA_NA;
        if(_outbuf.type & FLIT_DATA) {
            cmd_send.size -= 4;
            switch(cmd_send.cmd) {
            case 1://DMA_PUT
                cmd_send.remote_addr += cmd_send.remote_stride;
                break;
            case 2: //DMA_PUT64
                *(uint064_t*)&cmd_send.remote_addr += 4;
                break;
            default:
                cmd_send.remote_addr += cmd_send.remote_stride;
                break;
            }
            *(uint064_t*)&cmd_send.local_addr += 4;
        }
    }

    // recv command register
    if(inbuf.type & FLIT_VALID) {
        if(inbuf.type & FLIT_HEADER)
            cmd_recv.cmd = 0;
        else if(inbuf.type & FLIT_ADDR && cmd_recv.cmd == 0) {
            cmd_recv.local_addr = inbuf.body.data;
            cmd_recv.cmd = 1;
        }
        else if(inbuf.type & FLIT_ADDR && cmd_recv.cmd == 1) {
            cmd_recv.local_stride = inbuf.body.data;
        }
    }
    
    // main memory
    if(inbuf.type & FLIT_VALID && inbuf.type & FLIT_DATA) {
        mem->write4b(*(uint064_t*)&cmd_recv.local_addr,
                     inbuf.body.data);
        *(uint064_t*)&cmd_recv.local_addr += 4;
    }

    // dma_send_state
#if DMA_GET_BUF
    if(dma_send_state_ == INIT || router->ordy_[C])
#else
    if((dma_send_state_ == INIT || router->ordy_[C]) &&
       (dma_send_state_ != GET_FD || readable))
#endif
        dma_send_state = _dma_send_state;
    
    // dma_type
#if DMA_GET_BUF
    if(router->ordy_[C])
#else
    if(router->ordy_[C] && (dma_send_state_ != GET_FD || readable))
#endif
        dma_type = _dma_type;

    // inbuf
    inbuf = in;
    
    // rdy
    ordy = 1;
    irdy = router->ordy_[C];
    
    // send command register
    if(dma_send_state_ == INIT &&
       _dma_send_state != INIT) {
        if(! getreqbuf->isempty()) {
//            delete cmd_send.getdata;
            cmd_send = *(commcmd_t*)getreqbuf->head();
            getreqbuf->deq();
//            getdata->deq();
        }
    }
    
    // cmdbuf
    if(_getreqbuf.cmd) {
        core_incc_if[DMA_GET].cmd = 0;
        if(getreqbuf->isfull())
            nackmap->set(env->id2entry(_getreqbuf.remote_id));
        else {
/*
            _getreqbuf.getdata = new Getdata(
            *(uint064_t*)&_getreqbuf.local_addr,
                _getreqbuf.size, mem);
            getreqbuf->enq(&_getreqbuf);
*/
            getreqbuf->enq(&_getreqbuf);
            getdata->enq(*(uint064_t*)&_getreqbuf.local_addr,
                         _getreqbuf.size);
        }
    }
    
    // print
    if(env->log_mode == 1 || env->log_mode == 2)
        print_packet();
    if(env->log_mode == 2)
        print_flit();
}

/**********************************************************************/
/**
 * transfer flits between INCC64 and router,
 * calculate the next state
 */
void Incc64::comm() {
    const int lut[][2] = {
        {INIT},
        {INIT},
        {INIT, GET_FH},
        {INIT, ACK_FH},
        {INIT, NACK_FH}
    };
    
    nackentry = Bitmap::NOT_FOUND;
    
    // in
    in = router->out[C];
    
    // rdy
    ordy_ = ordy;
    irdy_ = irdy;

    // dma_send_state
    dma_send_state_ = dma_send_state;

    // dma_type
//    _dma_type = dma_type;
    dma_type_ = dma_type;
    
    // _getreqbuf
    if(core_incc_if[DMA_GET].cmd != DMA_NA)
        _getreqbuf = core_incc_if[DMA_GET];
    else
        _getreqbuf.cmd = 0;
    
    // _dma_send_state, _outbuf, _dma_send_state, _dma_type
    switch(dma_send_state_) {
    case INIT:
        if(! getreqbuf->isempty()) {
            _dma_send_state = lut[DMA_GET]
                [((commcmd_t*)getreqbuf->head())->cmd];
            _dma_type = DMA_GET;
            ready_counter = GMemory::READ_LATENCY;
//            printf("%llu %04x nack\n", env->cycle, _getreqbuf.remote_id);
        }
        else if(! nackmap->isclear()) {
            _dma_send_state = lut[DMA_NACK][1];
            _dma_type = DMA_NACK;
//            nackentry = nackmap->lsb();
//            printf("%llu nack ent:%d\n", env->cycle, nackentry);
//            nackmap->print();
        }
        break;
        
    case ACK_FH:
        _outbuf.type = FLIT_HEADER;
        _outbuf.header.coreid = cmd_send.remote_id;
        _outbuf.header.tag = 0;
        _dma_send_state = ACK_FA;
        break;
    case ACK_FA:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = SC_MMAP_DMA_FLAG;
        _dma_send_state = ACK_FD;
        break;
    case ACK_FD:
        _outbuf.type = FLIT_DATA | FLIT_TAILER;
        _outbuf.body.data = DMAFLAG_ACK;
        _dma_send_state = INIT;
        _dma_type = DMA_NA;
        break;

    case NACK_FH:
        _outbuf.type = FLIT_HEADER;
        nackentry = nackmap->lsb();
        nackmap->reset(nackentry);
//        nackentry = nackbuf->head()->remote_id;
//        nackbuf->deq();
        _outbuf.header.coreid = env->entry2id(nackentry);
        _outbuf.header.tag = 0;
        _dma_send_state = NACK_FA;
        break;
    case NACK_FA:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = SC_MMAP_DMA_FLAG;
        _dma_send_state = NACK_FD;
        break;
    case NACK_FD:
        _outbuf.type = FLIT_DATA | FLIT_TAILER;
        _outbuf.body.data = DMAFLAG_NACK;
        _dma_send_state = INIT;
        _dma_type = DMA_NA;
        break;

    case GET_FH:
        _outbuf.type = FLIT_HEADER;
        _outbuf.header.coreid = cmd_send.remote_id;
        _outbuf.header.tag = 0;
        _dma_send_state = GET_FA;
        break;
    case GET_FA:
        _outbuf.type = FLIT_ADDR;
        _outbuf.body.data = cmd_send.remote_addr;
        _dma_send_state = (cmd_send.remote_stride == 4) ?
            GET_FD : GET_FS;
        break;
    case GET_FS:
        _outbuf.type = FLIT_STRIDE;
        _outbuf.body.data = cmd_send.remote_stride;
        _dma_send_state = GET_FD;
        break;
#if DMA_GET_BUF
    case GET_FD:
        mem->read4b(*(uint064_t*)&cmd_send.local_addr, &_outbuf.body.data);
        if(cmd_send.size == 4) {
            _dma_send_state = ACK_FH;
            _dma_type = DMA_ACK;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else if(send_count == MAX_FLITS_PER_PACKET - 1) {
            _dma_send_state = GET_FH;
            _outbuf.type = (FLIT_DATA | FLIT_TAILER);
        } else {
            _dma_send_state = GET_FD;
            _outbuf.type = FLIT_DATA;
        }
        break;
#else
    case GET_FD:
        if(ready_counter > 0) {
            -- ready_counter;
            readable = 0;
            break;
        }
        
        readable = mem->pagehit(*(uint064_t*)&cmd_send.local_addr);
        
        if(readable) {
//            pagereadlady
            

//            mem->read4b(*(uint064_t*)&cmd_send.local_addr,
//              &_outbuf.body.data);
            _outbuf.body.data = getdata->nextdata();
            if(cmd_send.size == 4) {
                _dma_send_state = ACK_FH;
                _dma_type = DMA_ACK;
                _outbuf.type = (FLIT_DATA | FLIT_TAILER);
            } else if(send_count == MAX_FLITS_PER_PACKET - 1) {
                _dma_send_state = GET_FH;
                _outbuf.type = (FLIT_DATA | FLIT_TAILER);
            } else {
                _dma_send_state = GET_FD;
                _outbuf.type = FLIT_DATA;
            }
        }
        
        break;
#endif
    }
}

