/**********************************************************************/
/* Many-Core Architecture Research Project       Arch Lab. TOKYO TECH */
/**********************************************************************/
#include "network.h"

/**********************************************************************/
const char *Env::int2type(int type) {
    const char *lut[] = {
        "NODE", "SERVER", "MEM", "ROUTER"
    };
    return lut[type];
}

/**********************************************************************/
const char *Incc32::int2state(int state_no) {
    const char *lut[] = {
        "INIT     ",
        "ACK_FH   ", "ACK_FA   ", "ACK_FD   ",
        "NACK_FH  ", "NACK_FA  ", "NACK_FD  ", 
        "PUT_FH   ", "PUT_FA   ", "PUT_FS   ", "PUT_FD   ",
        "GET_FH   ", "GET_FA   ", "GET_FS   ", "GET_FD   ",
        "PUT64_FH ", "PUT64_FAL", "PUT64_FAH", "PUT64_FD ",
/*
        "GET0_FH  ", "GET0_FA  ", "GET0_FS  ", "GET0_FD  ",
        "GET1_FH  ", "GET1_FA  ", "GET1_FS  ", "GET1_FD  ",
        "GET2_FH  ", "GET2_FA  ", "GET2_FS  ", "GET2_FD  ",
        "GET3_FH  ", "GET3_FA  ", "GET3_FS  ", "GET3_FD  ",
*/
    };
    return ((state_no >= NUM_STATE) ?
            "UNDEFINED" : (char*)lut[state_no]);
}

/**********************************************************************/
const char *Incc32::int2dmatype(int dma_type) {
/*
    const char *_lut[][4] = {
        {"na    ",},
        {"na    ", "put   ", "put64 ", "puts  "},
        {"na    ", "get   ",},
        {"na    ", "ack   ",},
        {"na    ", "nack   ",},
    };
*/

    const char *lut[] = {
        "na   ",
        "put  ",
        "get  ",
        "ack  ",
        "nack ",
/*
        "get0 ",
        "get1 ",
        "get2 ",
        "get3 ",
*/
    };
    return (dma_type >= DMA_NUM_CMD ? "UNDEF" : lut[dma_type]);
}

/**********************************************************************/
const char *Incc64::int2dmatype(int dma_type) {
    const char *lut[] = {
        "na    ",
        "get64 ",
        "ack64 ",
        "nack64",
    };
    return (dma_type >= DMA_NUM_CMD ? "UNDEF" : lut[dma_type]);
}

void Env::print_log4(){
    for (int i = 0; i < nr_node; i++) {
        for (int j = 0; j < NUM_DIRECTION; j++) {
            if (router[i].in[j].type & FLIT_VALID && j != C) {
                log_t log = {cycle, phase,
                             (router[i].id & 0xff), (router[i].id >> 8), j, 0,
                             router[i].in[j].body.data & 0xff};
                fwrite(&log, sizeof(log_t), 1, fp);
            }
            if (!router[i].in_buf[j].isempty()) {
                log_t log = {cycle, phase,
                             (router[i].id & 0xff), (router[i].id >> 8), j, 1,
                             router[i].in_buf[j].head()->body.data & 0xff};
                fwrite(&log, sizeof(log_t), 1, fp);
            }
            if (router[i].out[j].type & FLIT_VALID && j != C) {
                log_t log = {cycle, phase,
                             (router[i].id & 0xff), (router[i].id >> 8), j, 2,
                             router[i].out[j].body.data & 0xff};
                fwrite(&log, sizeof(log_t), 1, fp);
            }
        }
    }
    if(phase++ == 2)
        phase = 0;
}

inline void const_log5(log5_t* l, int clk, flit_t* f, Router* r, int d, int s){
  l->clk = clk;
  l->type = (uint008_t)(f->type & 0xff);
  l->y = (uint008_t)(r->id & 0xff);
  l->x = (uint008_t)((r->id >> 8) & 0xff);
  l->dir_and_site = (uint008_t)(((d << 4) | s) & 0xff);
  l->data = f->body.data;
  /*
  printf("%08x [%d]: @(%d,%d):#(%d,%d), %08x", clk, f->type, (r->id & 0xff), ((r->id >> 8) & 0xff), d, s, f->body.data);
  printf(" ");
  unsigned char *p = (unsigned char*)l;
  for(int i = 0; i < 12; i++){
    printf("%02x ", *p);
    p++;
  }
  printf("\n");
  */
  return;
}

void Env::print_log5(){
  if(phase == 2){
    for (int i = 0; i < nr_node; i++) {
      for (int j = 0; j < NUM_DIRECTION; j++) {
	//if (router[i].in[j].type & FLIT_VALID && j != C) {
	/*
	if (router[i].in[j].type & FLIT_VALID) {
	  log5_t log;
	  const_log5(&log, cycle, &router[i].in[j], &router[i], j, 0);
	  fwrite(&log, sizeof(log5_t), 1, fp);
	}
	*/
	if (!router[i].in_buf[j].isempty()) {
	  log5_t log;
	  const_log5(&log, cycle, router[i].in_buf[j].head(), &router[i], j, 1);
	  fwrite(&log, sizeof(log5_t), 1, fp);
	}
	//if (router[i].out[j].type & FLIT_VALID && j != C) {
	if (router[i].out[j].type & FLIT_VALID) {
	  log5_t log;
	  const_log5(&log, cycle, &router[i].out[j], &router[i], j, 2);
	  fwrite(&log, sizeof(log5_t), 1, fp);
	}
      }
    }
    phase = 0;
  }else{
    phase++;
  }
}

/**********************************************************************/
void Env::print_log()
{
    if(log_mode == 4){
        print_log4();
    }else if(log_mode == 5) {
        print_log5();
    }else{
        return;
    }
}

/**********************************************************************/
static int print_1st;
void Incc32::print_packet()
{
    if(_dma_type != dma_type_ && _dma_type != DMA_NA) {
        if(!print_1st)
            puts(LOG1_HEADER), print_1st = 1;
        
        printf("%8lld %6s %04x %04x 0x%08x 0x%08x %8d %6d %6d\n",
               env->cycle, int2dmatype(cmd_send.cmd),
               *coreid, cmd_send.remote_id,
               cmd_send.local_addr, cmd_send.remote_addr,
               cmd_send.size,
               cmd_send.local_stride, cmd_send.remote_stride);
    }
}

/**********************************************************************/
void Incc64::print_packet()
{
    /*
    if(_dma_send_state == INIT || dma_send_state_ != INIT)
        return ;
    */    

    if(_dma_type != dma_type_ && _dma_type != DMA_NA) {
        if(!print_1st)
            puts(LOG1_HEADER), print_1st = 1;
        
        printf("%8lld %6s %04x %04x 0x%016llx 0x%08x %8d %6d\n",
               env->cycle, int2dmatype(cmd_send.cmd),
               *coreid, cmd_send.remote_id,
               *(uint064_t*)&cmd_send.local_addr, 
               cmd_send.remote_addr,
               cmd_send.size,
               cmd_send.remote_stride);
    }
}

void Env::print_flit(flit_t *flit) {
    if(flit->type & FLIT_VALID) {
        char type[6], *p = type;
        *p++ = (flit->type & FLIT_HEADER) ? 'h' : '-';
        *p++ = (flit->type & FLIT_ADDR) ? 'a' : '-';
        *p++ = (flit->type & FLIT_STRIDE) ? 's' : '-';
        *p++ = (flit->type & FLIT_DATA) ? 'd' : '-';
        *p++ = (flit->type & FLIT_TAILER) ? 't' : '-';
        *p = '\0';
        printf("  %6lld %s   0x%08x\n",
               cycle, type,
               (flit->type & FLIT_HEADER) ?
               ((flit->header.coreid << 16) | flit->header.tag) :
               flit->body.data);
    }
}

/**********************************************************************/
void Incc32::print_flit() {
    if(inbuf.type & FLIT_VALID) {
        char type[6], *p = type;
        *p++ = (inbuf.type & FLIT_HEADER) ? 'h' : '-';
        *p++ = (inbuf.type & FLIT_ADDR) ? 'a' : '-';
        *p++ = (inbuf.type & FLIT_STRIDE) ? 's' : '-';
        *p++ = (inbuf.type & FLIT_DATA) ? 'd' : '-';
        *p++ = (inbuf.type & FLIT_TAILER) ? 't' : '-';
        *p = '\0';
        printf("  %6lld recv    %s    %04x        0x%08x\n",
               env->cycle, type, *coreid,
               (inbuf.type & FLIT_HEADER) ?
               ((inbuf.header.coreid << 16) | inbuf.header.tag) :
               inbuf.body.data);
    }

    if(outbuf.type & FLIT_VALID) {
        char type[6], *p = type;
        *p++ = (_outbuf.type & FLIT_HEADER) ? 'h' : '-';
        *p++ = (_outbuf.type & FLIT_ADDR) ? 'a' : '-';
        *p++ = (_outbuf.type & FLIT_STRIDE) ? 's' : '-';
        *p++ = (_outbuf.type & FLIT_DATA) ? 'd' : '-';
        *p++ = (_outbuf.type & FLIT_TAILER) ? 't' : '-';
        *p = '\0';
        printf("  %6lld send    %s    %04x   %04x 0x%08x\n",
               env->cycle, type, *coreid, cmd_send.remote_id,
               (_outbuf.type & FLIT_HEADER) ?
               ((_outbuf.header.coreid << 16) | _outbuf.header.tag) :
               _outbuf.body.data);
    }
}

/**********************************************************************/
void Incc64::print_flit() {
    if(inbuf.type & FLIT_VALID) {
        char type[6], *p = type;
        *p++ = (inbuf.type & FLIT_HEADER) ? 'h' : '-';
        *p++ = (inbuf.type & FLIT_ADDR) ? 'a' : '-';
        *p++ = (inbuf.type & FLIT_STRIDE) ? 's' : '-';
        *p++ = (inbuf.type & FLIT_DATA) ? 'd' : '-';
        *p++ = (inbuf.type & FLIT_TAILER) ? 't' : '-';
        *p = '\0';
        printf("  %6lld recv    %s    %04x        0x%08x\n",
               env->cycle, type, *coreid,
               (inbuf.type & FLIT_HEADER) ?
               ((inbuf.header.coreid << 16) | inbuf.header.tag) :
               inbuf.body.data);
    }

    if(outbuf.type & FLIT_VALID) {
        char type[6], *p = type;
        *p++ = (_outbuf.type & FLIT_HEADER) ? 'h' : '-';
        *p++ = (_outbuf.type & FLIT_ADDR) ? 'a' : '-';
        *p++ = (_outbuf.type & FLIT_STRIDE) ? 's' : '-';
        *p++ = (_outbuf.type & FLIT_DATA) ? 'd' : '-';
        *p++ = (_outbuf.type & FLIT_TAILER) ? 't' : '-';
        *p = '\0';
        printf("  %6lld send    %s    %04x   %04x 0x%08x\n",
               env->cycle, type, *coreid, cmd_send.remote_id,
               (_outbuf.type & FLIT_HEADER) ?
               ((_outbuf.header.coreid << 16) | _outbuf.header.tag) :
               _outbuf.body.data);
    }
}

/**********************************************************************/
const char *Incc64::int2state(int state_no) {
    const char *lut[] = {
        "INCC64_INIT",
        "INCC64_GET_FH", "INCC64_GET_FA", "INCC64_GET_FS", "INCC64_GET_FD",
        "NUM_INCC64_STATE"

    };
    return ((state_no >= Incc64::NUM_STATE) ?
            "UNDEFINED" : (char*)lut[state_no]);
}

