/**********************************************************************/
/* Many-Core Architecture Research Project       Arch Lab. TOKYO TECH */
/**********************************************************************/
#include "network.h"

/**********************************************************************/
/**
 * Router constructor
 */
Router::Router()
{
    memset(last_send, 0, NUM_DIRECTION);
    memset(out, 0, NUM_DIRECTION);
    memset(in, 0, NUM_DIRECTION);
    memset(ordy, 1, NUM_DIRECTION);
}

/**********************************************************************/
/**
 * Router destructor
 */
Router::~Router() {}

/**********************************************************************/
/**
 * Router initialize
 * \param env Env pointer
 * \param id node id of the router
 */
void Router::init(Env *env, int id)
{
    int left, top, bottom, right;

    this->env = env;
    this->id = id;

    if(env->add_module) {
        left = 0;
        top = 0;
        right = env->noc_col - 1;
        bottom = env->noc_row - 1;
    }
    else {
        left = 1;
        top = 1;
        right = env->noc_col;
        bottom = env->noc_row;
    }

    router[N] = (IDY(id) == top) ? NULL :
        &env->router[env->id2entry(id) - env->noc_col];
    router[E] = (IDX(id) == right) ? NULL :
        &env->router[env->id2entry(id) + 1];
    router[S] = (IDY(id) == bottom) ? NULL :
        &env->router[env->id2entry(id) + env->noc_col];
    router[W] = (IDX(id) == left) ? NULL :
        &env->router[env->id2entry(id) - 1];
    
    incc = env->node[env->id2entry(id)].incc;

    do_round_robin[N] = (IDY(id) != top);
    do_round_robin[E] = (IDX(id) != right);
    do_round_robin[S] = (IDY(id) != bottom);
    do_round_robin[W] = (IDX(id) != left);
    do_round_robin[C] = 1;
    
    irdy[N] = (IDY(id) != top);
    irdy[E] = (IDX(id) != right);
    irdy[S] = (IDY(id) != bottom);
    irdy[W] = (IDX(id) != left);
    irdy[C] = 1;
}

/**********************************************************************/
/**
 * update registars
 */
void Router::update()
{
    // input buffer and do_round_robin
    for (int i = 0; i < NUM_DIRECTION; i++)
        if (out[i].type & FLIT_VALID) {
            if(in_buf[last_send[i]].deq() == false)
                printf("%llu: router(%d,%d) cross:%d->%d deq failed\n",
                       env->cycle, id >> 8, id & 0xff, last_send[i], i);

            if(out[i].type & FLIT_TAILER)
                do_round_robin[i] = 1;
            else if(out[i].type & FLIT_HEADER)
                do_round_robin[i] = 0;
        }
    
    // input buffer
    for (int i = 0; i < NUM_DIRECTION; i++)
        if (in[i].type & FLIT_VALID)
            in_buf[i].enq(in[i]);
    
    // ordy
    for (int i = 0; i < NUM_DIRECTION; i++)
        ordy[i] = (in_buf[i].num_entry() < Inbuf::BUF_SIZE - 1);
    
    // irdy
    irdy[C] = (incc == NULL ? 0 : incc->ordy_);
    for (int i = 0; i < NUM_DIRECTION - 1; i++)
        if (router[i] != NULL)
            irdy[i] = router[i]->ordy_[opposite(i)];
}

/**********************************************************************/
/**
 * routing computation, switch allocation and switch traversal
 */
void Router::switching()
{
    memcpy(ordy_, ordy, sizeof(uint008_t) * NUM_DIRECTION);

    // routing computation
    for (int i = 0; i < NUM_DIRECTION; i++)
        in_buf[i].rtcomp(id);

    // switch allocation
    uint008_t grant[NUM_DIRECTION];
    for (int i = 0; i < NUM_DIRECTION; i++) {
        grant[i] = 0;
        if (do_round_robin[i] && irdy[i]) {
            round_robin(j, last_send[i], NUM_DIRECTION) {
                if (in_buf[j].request == (1 << i)) {
                    last_send[i] = j;
                    grant[i] = 1;
                    break;
                }
            }
        }
    }

    // switch traversal
    for (int i = 0; i < NUM_DIRECTION; i++) {
        if(grant[i] ||
           (irdy[i] && !in_buf[last_send[i]].isempty() &&
            !do_round_robin[i]))
            out[i] = *in_buf[last_send[i]].head();
        else
//            out[i].type &= ~FLIT_VALID;
            out[i].type = 0;
    }
}

/**********************************************************************/
/**
 * transfar flits between routers
 */
void Router::comm()
{
    for (int i = 0; i < NUM_DIRECTION - 1; i++)
        if (router[i] != NULL)
            in[i] = router[i]->out[opposite(i)];

    in[C] = (incc == NULL ? flit_invalid : incc->outbuf);
}

